﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Problems.Instances.DataAnalysis.Regression.Matlab.Api.Types;


/* MLApp
 * https://www.codeproject.com/Articles/594636/Using-Matlab-from-a-Csharp-application
 *  string Execute(string Name)
 *  void Feval(string bstrName, int nargout, out object pvarArgOut, [object arg1], [object arg2], [object arg3], [object arg4], [object arg5], [object arg6], [object arg7], [object arg8], [object arg9], [object arg10], [object arg11], [object arg12], [object arg13], [object arg14], [object arg15], [object arg16], [object arg17], [object arg18], [object arg19], [object arg20], [object arg21], [object arg22], [object arg23], [object arg24], [object arg25], [object arg26], [object arg27], [object arg28], [object arg29], [object arg30], [object arg31], [object arg32])
 *  string GetCharArray(string Name, string Workspace)
 *  void GetFullMatrix(string Name, string Workspace, ref System.Array pr, ref System.Array pi)
 *  dynamic GetVariable(string Name, string Workspace)
 *  void GetWorkspaceData(string Name, string Workspace, out object pdata)
 *  void MaximizeCommandWindow()
 *  void MinimizeCommandWindow()
 *  void PutCharArray(string Name, string Workspace, string charArray)
 *  void PutFullMatrix(string Name, string Workspace, System.Array pr, System.Array pi)
 *  void PutWorkspaceData(string Name, string Workspace, object data)
 *  void Quit()
 *  
 * MLEval
 *  void XLEval(string bstrName, int nargout, ref object pvarArgOut, int nargin, object varArgIn)
 * */
namespace HeuristicLab.Problems.Instances.DataAnalysis.Regression.Matlab.Api {
  public class MatlabConnector : IMatlabConnector {
    private MLApp.MLApp matLabApi;
    private string workspace;
    private object locker = new object();

    public event Action<string> ErrorOccuredEvent;

    public MatlabConnector(bool showcommandWindow = false, string workspace = "base") {
      this.workspace = workspace;
      OpenMatlab(showcommandWindow);
    }

    private void OpenMatlab(bool showcommandWindow = false) {
      Type matlabtype = Type.GetTypeFromProgID("matlab.application.single");
      matLabApi = (MLApp.MLApp)Activator.CreateInstance(matlabtype);
      matLabApi.Visible = showcommandWindow ? 1 : 0;
    }

    public void ExecuteScript(string script) {
      string directoryName = Path.GetDirectoryName(script);
      if (string.IsNullOrEmpty(directoryName)) {
        directoryName = Environment.CurrentDirectory;
      }

      ChangeDirectory(directoryName);
      var result = matLabApi.Execute(Path.GetFileNameWithoutExtension(script));
      if (!string.IsNullOrEmpty(result)) {
        throw new InvalidOperationException(result);
      }
    }

    private void ChangeDirectory(string directoryName) {
      var result = matLabApi.Execute(string.Format("cd '{0}'", directoryName));
      if (!string.IsNullOrEmpty(result)) {
        throw new InvalidOperationException(result);
      }
    }

    public void Execute(string command) {
      matLabApi.Execute(command);
    }


    public IEnumerable<MLVariableInfo> GetVariablesFromScript(string script) {
      var variables = new List<MLVariableInfo>();
      try {
        ExecuteScript(script);
      } catch(InvalidOperationException ex) {
        if (ErrorOccuredEvent != null) {
          ErrorOccuredEvent.Invoke(ex.Message);
        }
      }

      Execute("variables=whos();nrOfVariables=size(variables)");
      var len = ((double[,])GetVariable("nrOfVariables"))[0, 0];

      for (int i = 1; i <= len; i++) {
        var varName = GetVariable(string.Format("variables({0}).name", i)) as string;
        var varSize = GetVariable(string.Format("variables({0}).size", i)) as double[,];
        var varClass = GetVariable(string.Format("variables({0}).class", i)) as string;
        
        GetVariablesFromScriptRec(variables, varName, varSize, varClass);
      }

      return variables;
    }

    private void GetVariablesFromScriptRec(IList<MLVariableInfo> variables, string varName, double[,] varSize, string varClass) {
      var variable = new MLVariableInfo() {
        VarName = varName,
        Row = varSize[0, 0],
        Column = varSize[0, 1]
      };
      if (varClass == "timeseries") {
        variable.Datatype = MLDatatype.Timeseries;
        variables.Add(variable);
      } else if (varClass == "double") {
        if (variable.Row == 1 && variable.Column == 1) {
          variable.Datatype = MLDatatype.Double;
        } else {
          variable.Datatype = MLDatatype.DoubleArray;
        }
        variables.Add(variable);
      } else if (varClass == "struct") {
        Execute(string.Format("mlFieldNames = fieldnames({0})", varName));
        var fields = GetVariable("mlFieldNames") as object[,];

        for (int i = 0; i < fields.GetLength(0); i++) {
          var newVarName = string.Format("{0}.{1}", varName, fields[i, 0] as string);
          Execute(string.Format("mlClass = class({0}); mlSize = size({0});", newVarName));
          var newVarClass = GetVariable("mlClass") as string;
          var newVarSize = GetVariable("mlSize") as double[,];
          GetVariablesFromScriptRec(variables, newVarName, newVarSize, newVarClass);
        }
      }
    }

    public MLTimeseries GetTimeseries(string varName) {
      lock (locker) {
        object times = GetVariable(varName + ".Time");
        object data = GetVariable(varName + ".Data");

        if (times == null || data == null) {
          return null;
        }
        return new MLTimeseries(varName, times, data);
      }
    }

    public MLDouble GetDouble(string varName) {
      return new MLDouble(varName, GetVariable(varName));
    }


    public MLDoubleArray GetDoubleArray(string varName) {
      return new MLDoubleArray(varName, GetVariable(varName));
    }

    #region Helper for getting the values of variables from the matlab api.        
    public object GetVariable(string varName) {
      object value = null;
      try {
        if (IsStruct(varName)) {
          value = GetValueOfStruct(varName);
        } else {
          matLabApi.GetWorkspaceData(varName, workspace, out value);
        }
      } catch (COMException) {
        throw new ArgumentException(string.Format("The variable {0} does not exist in the current workspace {1}.", varName, workspace));
      }
      var t = value.GetType();
      return value;
    }


    private bool IsStruct(string varName) {
      return varName.Contains(".");
    }

    private object GetValueOfStruct(string varName) {
      object value;
      matLabApi.Execute(string.Format("tmp = {0}", varName));
      matLabApi.GetWorkspaceData("tmp", workspace, out value);
      return value;
    }
    #endregion

    public void Dispose() {
      try {
        if (matLabApi != null) {
          matLabApi.Execute("exit");
        }
      } catch (COMException) {
      } finally {
        matLabApi = null;
      }
    }
  }
}

